from tornado.web import HTTPError
from biothings.web.api.es.handlers.base_handler import BaseESRequestHandler
from biothings.utils.web import sum_arg_dicts
import logging

class MetadataHandler(BaseESRequestHandler):
    ''' Request handlers for requests to the metadata endpoint. '''

    def initialize(self, web_settings):
        ''' Tornado handler `.initialize() <http://www.tornadoweb.org/en/stable/web.html#tornado.web.RequestHandler.initialize>`_ function for all requests to the metadata endpoint.
        Here, the allowed arguments are set (depending on the request method) for each kwarg category.'''
        super(MetadataHandler, self).initialize(web_settings)
        if self.request.method == 'GET':
            self.control_kwargs = self.web_settings.METADATA_GET_CONTROL_KWARGS
            self.es_kwargs = self.web_settings.METADATA_GET_ES_KWARGS
            self.esqb_kwargs = self.web_settings.METADATA_GET_ESQB_KWARGS
            self.transform_kwargs = self.web_settings.METADATA_GET_TRANSFORM_KWARGS
            self.kwarg_settings = sum_arg_dicts(self.control_kwargs, self.es_kwargs,
                                                self.esqb_kwargs, self.transform_kwargs)
        logging.debug("MetadataHandler - {}".format(self.request.method))
        logging.debug("Kwarg settings: {}".format(self.kwarg_settings))

    async def get(self):
        ''' Handle a GET to the metadata endpoint.  Also handles /metadata/fields. '''
        kwargs = self.get_query_params()

        if kwargs is None:
            return

        options = self.get_cleaned_options(kwargs)

        logging.debug("Request kwargs: {}".format(kwargs))
        logging.debug("Request options: {}".format(options))

        options = self._pre_query_builder_GET_hook(options)

        # Instantiate query builder, query and transform classes
        _query_builder = self.web_settings.ES_QUERY_BUILDER(
            options=options.esqb_kwargs,
            index=self._get_es_index(options),
            doc_type=self._get_es_doc_type(options),
            es_options=options.es_kwargs)
        _backend = self.web_settings.ES_QUERY(
            client=self.web_settings.async_es_client,
            options=options.es_kwargs)
        _result_transformer = self.web_settings.ES_RESULT_TRANSFORMER(
            options=options.transform_kwargs, host=self.request.host,
            app_dir=self.web_settings._app_git_repo,
            excluded_keys=self.web_settings.AVAILABLE_FIELDS_EXCLUDED,
            field_notes=self.web_settings.available_fields_notes())

        # get the query for annotation GET handler
        #try:
        _query = _query_builder.metadata_query()
        #except Exception as e:
        #    self.log_exceptions("Error building metadata query")
        #    self.return_json({'success': False, 'error': 'Error building query'})
        #    return

        logging.debug("Request query kwargs: {}".format(_query))

        # return raw query, if requested
        if options.control_kwargs.rawquery:
            self.return_object({}, rawquery=True, _format=options.control_kwargs.out_format)
            return

        _query = self._pre_query_GET_hook(options, _query)

        #try:
        res = await _backend.metadata_query(_query)
        #except Exception:
        #    self.log_exceptions("Error running query")
        #    self.return_json({'success': False, 'error': 'Error executing query'})
        #    return

        #logging.debug("Raw query result: {}".format(res))

        # return raw result if requested
        if options.control_kwargs.raw:
            self.return_object(res, _format=options.control_kwargs.out_format)
            return

        res = self._pre_transform_GET_hook(options, res)

        # clean result
        #try:
        res = _result_transformer.clean_metadata_response(
            res, fields=self.request.path.endswith('fields'))
        #except Exception:
        #    self.log_exceptions("Error transforming result")
        #    self.return_json({'success': False, 'error': 'Error transforming query result'})
        #    return

        res = self._pre_finish_GET_hook(options, res)

        self.return_object(res, _format=options.control_kwargs.out_format)



import json, jsonschema
from biothings.utils.common import is_str
from pprint import pprint,pformat
import bson


def generate_json_schema(dmap):

    scalarmap = {
            str: "string",
            int: "integer",
            float: "number",
            bool: "boolean",
            bson.int64.Int64: "number",
            None: "null",
            }

    def merge_type(typ1,typ2):
        if type(typ1) == list:
            if type(typ2) == list:
                typ1.extend(typ2)
            else:
                typ1.append(typ2)
        elif type(typ2) == list:
            typ1 = [typ1] + typ1
        else:
            typ1 = [typ1,typ2]

        return list(set(typ1))

    schema = {}

    if type(dmap) == dict:
        for k in dmap:
            if is_str(k):
                esch = generate_json_schema(dmap[k])
                if schema:
                    if schema["type"] == "object":
                        # we just complete 'properties', key already defined previously
                        pass
                    elif schema["type"] == "array":
                        if not schema.get("properties"):
                            schema["properties"] = {}
                            schema["type"] = merge_type(schema["type"],"object")
                    elif type(schema["type"]) == list:
                        assert set(schema["type"]) == {"object","array"}
                        pass
                    else:
                        raise Exception("Previous schema type not expected: %s" % schema["type"])

                else:
                    schema = {"type" : "object", "properties" : {}}
                schema["properties"][k] = esch
            elif type(k) == type:
                if k == list:
                    if schema:
                        # already defined for this key, mixed types
                        schema.update({"items" : {}})
                        schema["type"] = merge_type(schema["type"],"array")
                    else:
                        schema = {"type" : "array", "items" : {}}
                    esch = generate_json_schema(dmap[k])
                    schema["items"] = generate_json_schema(dmap[k])
                else:
                    if schema:
                        schema["type"] = merge_type(schema["type"],scalarmap[k])
                    else:
                        schema = {"type" : scalarmap[k]}
            elif k == None:
                schema = {"type" : None}
            else:
                raise Exception("no not here, k: %s" % k)
    else:
        pass

    return schema


def test():
    from biothings.utils.inspect import typify_inspect_doc, inspect_docs
    import pickle
    from pprint import pprint,pformat

    # can't use assert directly, as we can't ensure the order of types (for instance)
    import biothings.utils.jsondiff
    biothings.utils.jsondiff.UNORDERED_LIST = True
    jsondiff = biothings.utils.jsondiff.make

    # object
    td1 = {"i" : {"a":456}}
    s1 = {'properties': {'i': {'properties': {'a': {'type': 'integer'}},
       'type': 'object'}},
     'type': 'object'}
    m = inspect_docs([td1],mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs,s1) == [], "%s  !=\n%s" % (gs,s1)

    td5 = {"i" : [1,2,3]}
    s5 = {'properties': {'i': {'items': {'type': 'integer'}, 'type': 'array'}},
            'type': 'object'}
    m = inspect_docs([td5],mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs,s5) == [], "%s  !=\n%s" % (gs,s5)

    # array of object
    td2 = {"i" : [{"a":123}]}
    s2 = {'properties': {'i': {'items': {'properties': {'a': {'type': 'integer'}},
        'type': 'object'},
       'type': 'array'}},
     'type': 'object'}
    m = inspect_docs([td2],mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs,s2) == [], "%s  !=\n%s" % (gs,s2)

    # object in object
    td3 = {"i" : {"a":{"b":123}}}
    s3 = {'properties': {'i': {'properties': {'a': {'properties': {'b': {'type': 'integer'}},
         'type': 'object'}},
       'type': 'object'}},
     'type': 'object'}
    m = inspect_docs([td3],mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs,s3) == [], "%s  !=\n%s" % (gs,s3)

    # mixed str/float in array
    td6 = {"i" : [1,2,"a"]}
    s6 = {'properties': {'i': {'items': {'type': ['integer','string']}, 'type': 'array'}},
            'type': 'object'}
    m = inspect_docs([td6],mode="type")["type"]
    gs = generate_json_schema(m) 
    assert jsondiff(gs,s6) == [], "%s  !=\n%s" % (gs,s6)

    # mixed array/object
    td1 = {"i" : {"a":456}}
    td2 = {"i" : [{"a":123}]}
    s12 = {'properties': {'i': {'items': {'properties': {'a': {'type': 'integer'}},
        'type': 'object'},
       'properties': {'a': {'type': 'integer'}},
       'type': ['array', 'object']}},
     'type': 'object'}
    m = inspect_docs([td1,td2],mode="type")["type"] 
    gs = generate_json_schema(m)
    assert jsondiff(gs,s12) == [], "%s  !=\n%s" % (gs,s12)

    # list of integer (list of things which are not objects)
    td4 = {'a': [5, 5, 3]}
    s4 = {'properties': {'a': {'items': {'type': 'integer'},
        'type': 'array'}},
      'type': 'object'}
    m = inspect_docs([td4],mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs,s4) == [], "%s  !=\n%s" % (gs,s4)

    td7 = {"i" : {"a":1,"b":2}}
    s7 = {'type': 'object', 'properties': {
            'i': {'type': 'object', 'properties': {
                'a': {'type': 'integer'},
                'b': {'type': 'integer'}
          }}}}
    m = inspect_docs([td7],mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs,s7) == [], "%s  !=\n%s" % (gs,s7)

    # int or list of int (not a list of dict, testing scalar there)
    td81 = {"i": 1}
    td82 = {"i" : [2,3]}
    s812 = {'properties': {'i': {'items': {'type': 'integer'}, 'type': ['array', 'integer']}},
            'type': 'object'}
    m = inspect_docs([td81,td82],mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs,s812) == [], "%s  !=\n%s" % (gs,s812)

    # run from app folder, biothings as symlink

    # small real-life collection
    cgi_schema = json.load(open("biothings/tests/cgi_schema.json"))
    cgi_map = typify_inspect_doc(json.load(open("biothings/tests/cgi_map.json")))
    schema = generate_json_schema(cgi_map)
    assert jsondiff(cgi_schema,schema) == []

    clinvar_schema = json.load(open("biothings/tests/clinvar_schema.json"))
    clinvar_map = typify_inspect_doc(json.load(open("biothings/tests/clinvar_map.json")))
    schema = generate_json_schema(clinvar_map)
    assert jsondiff(clinvar_schema,schema) == []

    mygene_schema = json.load(open("biothings/tests/mygene_schema.json"))
    mygene_map = typify_inspect_doc(json.load(open("biothings/tests/mygene_map.json")))
    schema = generate_json_schema(mygene_map)
    assert jsondiff(mygene_schema,schema) == []

    print("All test OK")


if __name__ == "__main__":
    test()



''' Functions to return versions of things. '''
from subprocess import check_output, DEVNULL
from io import StringIO
from contextlib import redirect_stdout
# import pip
import os
import sys
import logging
from git import Repo, Git, GitCommandError, NoSuchPathError, InvalidGitRepositoryError
import re
import biothings
from biothings.utils.dataload import dict_sweep

# cache these
BIOTHINGS_REPO_DATA = {}
APP_REPO_DATA = {}


def get_python_version():
    ''' Get a list of python packages installed and their versions. '''
    try:
        return check_output('pip list',
                            shell=True, stderr=DEVNULL).decode('utf-8').split('\n')[2: -1]
    except Exception:
        return []


def get_biothings_commit():
    ''' Gets the biothings commit information. '''
    global BIOTHINGS_REPO_DATA
    if BIOTHINGS_REPO_DATA:
        return BIOTHINGS_REPO_DATA

    try:
        with open(os.path.join(os.path.dirname(biothings.__file__), '.git-info'), 'r') as f:
            lines = [l.strip('\n') for l in f.readlines()]
        BIOTHINGS_REPO_DATA = {
            'repository-url': lines[0],
            'commit-hash': lines[1],
            'master-commits': lines[2],
            'version': biothings.get_version()}
    except:
        BIOTHINGS_REPO_DATA = {'repository-url': '', 'commit-hash': '',
                               'master-commits': '', 'version': biothings.get_version()}

    return BIOTHINGS_REPO_DATA


def get_repository_information(app_dir=None):
    ''' Get the repository information for the local repository, if it exists. '''
    global APP_REPO_DATA
    if APP_REPO_DATA:
        return APP_REPO_DATA

    if not app_dir:
        APP_REPO_DATA = {'repository-url': '', 'commit-hash': ''}
        return APP_REPO_DATA

    try:
        commit_hash = check_output(
            "cd {};git rev-parse HEAD".format(os.path.abspath(app_dir)),
            shell=True, stderr=DEVNULL).decode('utf-8').strip('\n')
    except:
        commit_hash = ''

    try:
        repository_url = check_output(
            "cd {};git config --get remote.origin.url".format(os.path.abspath(app_dir)),
            shell=True, stderr=DEVNULL).decode('utf-8').strip('\n')
    except:
        repository_url = ''

    APP_REPO_DATA = {'repository-url': repository_url, 'commit-hash': commit_hash}

    return APP_REPO_DATA


def get_python_exec_version():
    return {
        'version': sys.version,
        'version_info': {
            "major": sys.version_info[0],
            "minor": sys.version_info[1],
            "micro": sys.version_info[2]
        }
    }


def get_software_info(app_dir=None):
    return {
        'python-package-info': get_python_version(),
        'codebase': get_repository_information(app_dir=app_dir),
        'biothings': get_biothings_commit(),
        'python-info': get_python_exec_version(),
    }


def set_versions(config, app_folder):
    """
    Propagate versions (git branch name) in config module
    """
    # app_version: version of the API application
    if not hasattr(config, "APP_VERSION"):
        repo = Repo(app_folder)  # app dir (mygene, myvariant, ...)
        try:
            commit = repo.head.object.hexsha[:6]
            commitdate = repo.head.object.committed_datetime.isoformat()
        except Exception as e:
            logging.warning("Can't determine app commit hash: %s" % e)
            commit = "unknown"
            commitdate = "unknown"
        try:
            config.APP_VERSION = {"branch" : repo.active_branch.name,
                                  "commit" : commit,
                                  "date" : commitdate}
        except Exception as e:
            logging.warning("Can't determine app version, defaulting to 'master': %s" % e)
            config.APP_VERSION = {"branch" : "master",
                                  "commit" : commit,
                                  "date" : commitdate}
    else:
        logging.info("app_version '%s' forced in configuration file" % config.APP_VERSION)

    # biothings_version: version of BioThings SDK
    if not hasattr(config, "BIOTHINGS_VERSION"):
        import biothings
        # .../biothings.api/biothings/__init__.py
        bt_folder, _bt = os.path.split(os.path.split(os.path.realpath(biothings.__file__))[0])
        assert _bt == "biothings", "Expectig 'biothings' dir in biothings lib path"
        repo = Repo(bt_folder)  # app dir (mygene, myvariant, ...)
        try:
            commit = repo.head.object.hexsha[:6]
            commitdate = repo.head.object.committed_datetime.isoformat()
        except Exception as e:
            logging.warning("Can't determine biothings commit hash: %s" % e)
            commit = "unknown"
            commitdate = "unknown"
        try:
            config.BIOTHINGS_VERSION = {"branch" : repo.active_branch.name,
                                        "commit" : commit,
                                        "date" : commitdate}
        except Exception as e:
            logging.warning("Can't determine biothings version, defaulting to 'master': %s" % e)
            config.BIOTHINGS_VERSION = {"branch" : "master",
                                        "commit" : commit,
                                        "date" : commitdate}
    else:
        logging.info("biothings_version '%s' forced in configuration file" %
                     config.BIOTHINGS_VERSION)

    logging.info("Running app_version=%s with biothings_version=%s" %
                 (repr(config.APP_VERSION), repr(config.BIOTHINGS_VERSION)))


def get_source_code_info(src_file):
    """
    Given a path to a source code, try to find information
    about repository, revision, URL pointing to that file, etc...
    Return None if nothing can be determined.
    Tricky cases: 
      - src_file could refer to another repo, within current repo
        (namely a remote data plugin, cloned within the api's plugins folder
      - src_file could point to a folder, when for instance a dataplugin is
        analized. This is because we can't point to an uploader file since
        it's dynamically generated
    """
    # need to be absolute to build proper github URL
    abs_src_file = os.path.abspath(src_file)
    try:
        repo = Repo(abs_src_file, search_parent_directories=True)
    except (InvalidGitRepositoryError, NoSuchPathError) as e:
        logging.exception("Can't find a github repository for file '%s'" % src_file)
        return None
    try:
        gcmd = Git(repo)
        hash = gcmd.rev_list(-1, repo.active_branch, abs_src_file)
        rel_src_file = abs_src_file.replace(repo.working_dir, "").strip("/")
        if not hash:
            # seems to be a repo cloned within a repo, change directory
            curdir = os.path.abspath(os.curdir)
            try:
                if os.path.isdir(abs_src_file):
                    os.chdir(abs_src_file)
                    hash = gcmd.rev_list(-1, repo.active_branch)
                else:
                    dirname, filename = os.path.split(abs_src_file)
                    os.chdir(dirname)
                    hash = gcmd.rev_list(-1, repo.active_branch, filename)
                rel_src_file = ""  # will point to folder by commit hash
            finally:
                os.chdir(curdir)
        if hash:
            short_hash = gcmd.rev_parse(hash, short=7)
        else:
            logging.warning("Couldn't determine commit hash for file '%s'" % src_file)
            hash = None
            short_hash = None
        # could have more than one URLs for origin, only take first
        repo_url = next(repo.remote().urls)
        info = {
            "repo": repo_url,
            "commit": short_hash,
            "branch": repo.active_branch.name,
        }
        if os.path.isdir(src_file):
            info["folder"] = rel_src_file
        else:
            info["file"] = rel_src_file
        info = dict_sweep(info)
        # rebuild URL to that file
        if "github.com" in repo_url:
            info["url"] = os.path.join(re.sub("\.git$", "", repo_url),
                                       "tree", hash, rel_src_file)

        return info

    except GitCommandError as e:
        logging.exception("Error while getting git information for file '%s'" % src_file)
        return None
    except TypeError as e:
        # happens with biothings symlink, just ignore
        logging.debug("Can't determine source code info (but that's fine): %s" % e)
        return None



import os
import sys
import tornado.httpserver
import tornado.ioloop
import tornado.web
import tornado.escape
from biothings.web.settings import BiothingESWebSettings
from biothings.utils.common import is_str
from tempfile import NamedTemporaryFile
from string import Template

def _is_module(m):
    try:
        import types
        if type(m) == types.ModuleType:
            return True
        import importlib
        _ret = importlib.import_module(m)
    except ImportError:
        return False

    return _ret

def _is_file(f):
    return os.path.exists(os.path.abspath(f)) and os.path.isfile(os.path.abspath(f))

class BiothingsAPIApp(object):
    def __init__(self, *args, **kwargs):
        if kwargs.get('object_name', False) or (len(args) >= 1 and not _is_module(args[0]) and not _is_file(args[0])):
            _arg = args[0] if args else ''
            self._configure_by_object_name(object_name=kwargs.get('object_name', _arg))
        elif ((kwargs.get('config_file', False) and _file_exists(kwargs['config_file'])) or
            (len(args) >= 1 and _is_file(args[0]))):
            _arg = args[0] if args else ''
            self._configure_by_file(config_file=os.path.abspath(kwargs.get('config_file', _arg)))
        elif ((kwargs.get('config_module', False) and _is_module(kwargs['config_module'])) or 
            (len(args) >= 1 and _is_module(args[0]))):
            _arg = args[0] if args else ''
            self._configure_by_module(config_module=kwargs.get('config_module', _arg))
        else:
            self._configure_by_kwargs(**kwargs)

    def get_server(self,config_mod, **app_settings):
        settings = BiothingESWebSettings(config=config_mod) 
        app = tornado.web.Application(settings.generate_app_list(), **app_settings)
        server = tornado.httpserver.HTTPServer(app)
        return server

    def start(self, debug=True, port=8000, address='127.0.0.1', app_settings={}):
        if debug:
            #import tornado.autoreload
            import logging
            logging.getLogger().setLevel(logging.DEBUG)
            address='0.0.0.0'

        with NamedTemporaryFile(mode='w', suffix='.py', dir=os.path.abspath('.')) as _tempfile:
            if not getattr(self, 'settings_mod', False):
                if not getattr(self, 'settings_str', False):
                    raise("Could not configure biothings app")
                _tempfile.file.write(self.settings_str)
                _tempfile.file.flush()
                self.settings_mod = os.path.split(_tempfile.name)[1].split('.')[0]
            self.settings = BiothingESWebSettings(config=self.settings_mod)
            application = tornado.web.Application(self.settings.generate_app_list(), **app_settings)
            http_server = tornado.httpserver.HTTPServer(application)
            http_server.listen(port, address)
            loop = tornado.ioloop.IOLoop.instance()
            if debug:
                #tornado.autoreload.start(loop)
                logging.info('Server is running on "%s:%s"...' % (address, port))
            loop.start() 

    def _configure_by_object_name(self, object_name):
        # get the config file template
        config_string = """from biothings.web.settings.default import *\n""" + \
                        """from biothings.web.api.es.handlers import *\n""" + \
                        """ES_INDEX = '${src_package}_current'\n""" + \
                        """ES_DOC_TYPE = '${es_doctype}'\n""" + \
                        """API_VERSION = 'v1'\n""" + \
                        """APP_LIST = [(r'/status', StatusHandler), (r'/metadata/?', MetadataHandler), (r'/metadata/fields/?', MetadataHandler), (r'/{}/${annotation_endpoint}/(.+)/?'.format(API_VERSION), BiothingHandler), (r'/{}/${annotation_endpoint}/?$$'.format(API_VERSION), BiothingHandler), (r'/{}/query/?'.format(API_VERSION), QueryHandler), (r'/{}/metadata/?'.format(API_VERSION), MetadataHandler), (r'/{}/metadata/fields/?'.format(API_VERSION), MetadataHandler),]\n""" + \
                        """GA_RUN_IN_PROD = False"""
        settings_dict = {'src_package': 'my' + object_name.lower(),
                         'es_doctype': object_name.lower(),
                         'annotation_endpoint': object_name.lower()}
       
        self.settings_str = Template(config_string).substitute(settings_dict)            

    def _configure_by_module(self, config_module):
        self.settings_mod = config_module

    def _configure_by_kwargs(self, **kwargs):
        self.settings_str = """from biothings.web.settings.default import *\nfrom biothings.web.api.es.handlers import *\n"""
        for (k,v) in kwargs.items():
            if k == 'APP_LIST':
                self.settings_str += '{k}=['.format(k=k)
                for (reg, handler_str) in v:
                    self.settings_str += "(r'{reg}', {handler}),".format(reg=reg, handler=handler_str)
                self.settings_str += ']\n'
            elif k in ['ES_QUERY_BUILDER', 'ES_QUERY', 'ES_RESULT_TRANSFORMER']:
                self.settings_str += '{k}={v}\n'
            elif is_str(v):
                self.settings_str += '{k}="{v}"\n'.format(k=k, v=v)
            else:
                self.settings_str += '{k}={v}\n'
        
    def _configure_by_file(self, config_file):
        with open(config_file, 'r') as config_handle:
            self.settings_str = config_handle.read()



import sys

from biothings.utils.hub_db import get_src_dump, get_data_plugin, get_hub_db_conn, backup, restore
from biothings import config
logging = config.logger

def migrate_0dot1_to_0dot2():
    """
    mongodb src_dump/data_plugin changed:
        1. "data_folder" and "release" under "download"
        2. "data_folder" and "release" in upload.jobs[subsrc] taken from "download"
        3. no more "err" under "upload"
        4. no more "status" under "upload"
        5. "pending_to_upload" is now "pending": ["upload"]
    """
    src_dump = get_src_dump()
    data_plugin = get_data_plugin()
    for srccol in [src_dump,data_plugin]:
        logging.info("Converting collection %s" % srccol)
        srcs = [src for src in srccol.find()]
        wasdue = False
        for src in srcs:
            logging.info("\tConverting '%s'" % src["_id"])
            # 1.
            for field in ["data_folder","release"]:
                if field in src:
                    logging.debug("\t\t%s: found '%s' in document, moving under 'download'" % (src["_id"],field))
                    try:
                        src["download"][field] = src.pop(field)
                        wasdue = True
                    except KeyError as e:
                        logging.warning("\t\t%s: no such field '%s' found, skip it (error: %s)" % (src["_id"],field,e))
            # 2.
            for subsrc_name in src.get("upload",{}).get("jobs",{}):
                for field in ["data_folder","release"]:
                    if not field in src["upload"]["jobs"][subsrc_name]:
                        logging.debug("\t\t%s: no '%s' found in upload jobs, taking it from 'download' (or from root keys)" % (src["_id"],field))
                        try:
                            src["upload"]["jobs"][subsrc_name][field] = src["download"][field]
                            wasdue = True
                        except KeyError:
                            try:
                                src["upload"]["jobs"][subsrc_name][field] = src[field]
                                wasdue = True
                            except KeyError:
                                logging.warning("\t\t%s: no such field '%s' found, skip it" % (src["_id"],field))
            # 3. & 4.
            for field in ["err","status"]:
                if field in src.get("upload",{}):
                    logging.debug("\t\t%s: removing '%s' key from 'upload'" % (src["_id"],field))
                    src["upload"].pop(field)
                    wasdue = True
            # 5.
            if "pending_to_upload" in src:
                logging.debug("\t%s: found 'pending_to_upload' field, moving to 'pending' list" % src["_id"])
                src.pop("pending_to_upload")
                wasdue = True
                if not "upload" in src.get("pending",[]):
                    src.setdefault("pending",[]).append("upload")
            if wasdue:
                logging.info("\tFinishing converting document for '%s'" % src["_id"])
                srccol.save(src)
            else:
                logging.info("\tDocument for '%s' already converted" % src["_id"])


def migrate(from_version, to_version,restore_if_failure=True):
    func_name = "migrate_%s_to_%s" % (from_version.replace(".","dot"),
                              to_version.replace(".","dot"))
    # backup
    db = get_hub_db_conn()[config.DATA_HUB_DB_DATABASE]
    logging.info("Backing up %s" % db.name)
    path = backup()
    logging.info("Backup file: %s" % path)
    thismodule = sys.modules[__name__]
    try:
        func = getattr(thismodule,func_name)
    except AttributeError:
        logging.error("Can't upgrade, no such function to migrate from '%s' to '%s'" % (from_version, to_version))
        raise
    # resolve A->C = A->B then B->C
    logging.info("Start upgrading from '%s' to '%s'" % (from_version, to_version))
    try:
        func()
    except Exception as e:
        logging.exception("Failed upgrading: %s")
        if restore_if_failure:
            logging.info("Now restoring original database from '%s" % path)
            restore(db,path,drop=True)
            logging.info("Done. If you want to keep converted data for inspection, use restore_if_failure=False")
        else:
            logging.info("*not* restoring original data. It can still be restored using file '%s'" % path)








''' A collection of html templates used by biothings APIs '''

HTML_OUT_TEMPLATE = '''
    <html>
        <head>
            <link href="https://cdn.rawgit.com/josdejong/jsoneditor/v5.11.0/dist/jsoneditor.min.css" rel="stylesheet" type="text/css">
            <script src="https://ajax.googleapis.com/ajax/libs/jquery/1.12.4/jquery.min.js"></script>
            <script src="https://cdn.rawgit.com/josdejong/jsoneditor/v5.11.0/dist/jsoneditor.min.js"></script>
            <script type="text/javascript">
                $(document).ready(function () {{
                    var container = document.getElementById("jsonview");
                    var options = {{
                        mode: 'view'
                    }};
                    var editor = new JSONEditor(container, options);
                    var data = {data};
                    editor.set(data);
                }});
            </script>
        </head>
        <body>
            <table style="width:100%; height:100%">
                <tbody>
                    <tr style="height:35px;">
                        <td style="width: 35px"><img src="{img_src}" alt="logo" height="35" width="35"></img></td>
                        <td style="vertical-align:center; padding-left:5px;">{title_html}</td>
                    </tr>
                    <tr style="height:15px;">
                        <td colspan="2"><p><a href="{link}" target="_blank">{link_decode}</a>&nbsp;&nbsp;[<a href="{docs_link}" target="_blank">Learn more about this API endpoint</a>]</p></td>
                    </tr>
                    <tr><td colspan="2"><div id="jsonview" style="overflow:auto; height: 100%"></div></td></tr>
                </tbody>
            </table>
        </body>
    </html>'''



import os.path
import re
#import logging

query_cache = {}
filter_cache = {}

def get_userquery(query_folder, query_name):
    try:
        return query_cache[query_name]
    except KeyError:
        pass

    # get the query from file, assumed that query_folder/query_name/query.txt exists
    query_dir = os.path.join(query_folder, query_name)
    with open(os.path.join(query_dir, 'query.txt'), 'r') as query_handle:
        query_cache[query_name] = re.sub(r'\{\{\{\{(?P<var>.*?)\}\}\}\}', '{\g<var>}', 
                                  re.sub(r"\{", "{{", re.sub(r"\}", "}}", query_handle.read())))
    #logging.debug("query_cache[{}]: {}".format(query_name, query_cache[query_name]))

    return query_cache[query_name]

def get_userfilter(query_folder, query_name):
    try:
        return filter_cache[query_name]
    except KeyError:
        pass

    # get the filter from file, assumed that query_folder/query_name/filter.txt exists
    query_dir = os.path.join(query_folder, query_name)
    with open(os.path.join(query_dir, 'filter.txt'), 'r') as filter_handle:
        filter_cache[query_name] = filter_handle.read()
    #logging.debug("filter_cache[{}]: {}".format(query_name, filter_cache[query_name]))

    return filter_cache[query_name]



from tornado.web import HTTPError
from biothings.web.api.es.handlers.base_handler import BaseESRequestHandler

class StatusHandler(BaseESRequestHandler):
    ''' Handles requests to check the status of the server. '''

    async def head(self):
        try:
            r = await self.web_settings.async_es_client.get(**self.web_settings.STATUS_CHECK)
        except BaseException:
            raise HTTPError(503)

        if not r:
            raise HTTPError(503)

    async def get(self):
        await self.head()
        self.write('OK')



# -*- coding: utf-8 -*-
'''
Nose tests
run as "nosetests tests"
    or "nosetests tests:test_main"
'''
from urllib.parse import urlencode
from urllib.parse import urlparse
from urllib.parse import quote_plus
import json
import os
import re
import sys
import unittest
from functools import wraps

import requests

from biothings.tests import BiothingsTestCase
from biothings.tests.web.settings import BiothingTestSettings
from biothings.tests.web.helper import parameterized, parameters

try:
    import msgpack
except ImportError:
    sys.stderr.write("Warning: msgpack is not available.")

bts = BiothingTestSettings(config_module='tests.test_config')

_d = json.loads    # shorthand for json decode
_e = json.dumps    # shorthand for json encode
_q = quote_plus     # shorthand for url encoding

try:
    jsonld_context = requests.get(bts.JSONLD_CONTEXT_URL).json()
except BaseException:
    sys.stderr.write("Couldn't load JSON-LD context.\n")
    jsonld_context = {}

PATTR = '%values'


@parameterized
class BiothingGenericTests(BiothingsTestCase):
    __test__ = False  # don't run nosetests on this class directly

    # Make these class variables so that tornadorequesthelper plays nice.
    host = os.getenv(bts.HOST_ENVAR_NAME, '')
    if not host:
        host = bts.NOSETEST_DEFAULT_URL
    host = host.rstrip('/')
    api = '/' + bts.API_VERSION

    # Setup/tear down class for unittest
    @classmethod
    def setUpClass(cls):
        # get host and urls, etc
        sys.stderr.write("Running tests on {}\n".format(cls.host))
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    # setup tear down test for unittest
    def setUp(self):
        pass

    def tearDown(self):
        pass

    #############################################################
    # Test functions                                            #
    #############################################################

    def test_main(self):
        self.request(self.host)

    #############################################################
    # Annotation endpoint GET tests                             #
    #############################################################

    def test_annotation_object(self):
        ''' Test that annotation object contains expected fields. '''
        url = bts.ANNOTATION_ENDPOINT + '/' + bts.ANNOTATION_OBJECT_ID
        res = self.request(url).json()
        for attr in bts.ANNOTATION_OBJECT_EXPECTED_ATTRIBUTE_LIST:
            self.assertIn(
                attr, res, 'Expected field "{}" in returned object of query: {}'.format(
                    attr, url))

    def test_annotation_GET_empty(self):
        ''' Test that the annotation GET endpoint handles empty inputs correctly. '''
        self.request('' + bts.ANNOTATION_ENDPOINT, expect_status=404)
        self.request('' + bts.ANNOTATION_ENDPOINT + '/', expect_status=404)

    def test_annotation_GET_unicode(self):
        ''' Test that the annotation GET endpoint handles unicode string inputs correctly. '''
        self.request(
            ''
            + bts.ANNOTATION_ENDPOINT
            + '/'
            + bts.UNICODE_TEST_STRING,
            expect_status=404)

    @parameters(bts.ANNOTATION_GET_FIELDS)
    def test_annotation_GET_fields(self, bid):
        ''' Test that the fields parameter on the annotation GET endpoint works as expected. '''
        url = bts.ANNOTATION_ENDPOINT + '/' + bid
        # Test setup correctly?
        self.assertRegexpMatches(
            bid, r'fields=.+',
            'Test setup error, "fields" parameter not specified or empty in url: {}'.format(url))
        # get the url with all fields
        total_url = bts.ANNOTATION_ENDPOINT + '/' + _q(bid.split('?')[0])
        # get the user specified fields
        true_fields = [x.strip() for x in self.parse_url(url, 'fields').split(',')]
        # get the original and total request (with specified fields, and all fields, respectively)
        res = self.request(url).json()
        res_total = self.request(total_url).json()
        # check the fields
        self.check_fields(res, res_total, true_fields,
                          bts.CHECK_FIELDS_SUBSET_ADDITIONAL_FIELDS)

    @parameters(bts.ANNOTATION_GET_FILTER)
    def test_annotation_GET_filter(self, bid):
        ''' Test that the filter parameter on the annotation GET endpoint works as expected. '''
        url = bts.ANNOTATION_ENDPOINT + '/' + bid
        # Test setup correctly?
        self.assertRegexpMatches(
            bid, r'filter=.+',
            'Test setup error, "filter" parameter not specified or empty in url: {}'.format(url))
        # get the url with all fields
        total_url = bts.ANNOTATION_ENDPOINT + '/' + _q(bid.split('?')[0])
        # get the user specified fields
        true_filter = [x.strip() for x in self.parse_url(url, 'filter').split(',')]
        # get the original and total request (with specified fields, and all fields, respectively)
        res = self.request(url).json()
        res_total = self.request(total_url).json()
        # check the fields
        self.check_fields(res, res_total, true_filter,
                          bts.CHECK_FIELDS_SUBSET_ADDITIONAL_FIELDS)

    @parameters(bts.ANNOTATION_GET_JSONLD)
    def test_annotation_GET_jsonld(self, bid):
        ''' Test that the annotation GET endpoint correctly returns the JSON-LD context information. '''
        url = bts.ANNOTATION_ENDPOINT + '/' + bid
        self.assertRegexpMatches(
            bid, r'jsonld=((1)|([Tt][Rr][Uu][Ee]))',
            'Test setup error, "jsonld" parameter not specified or false in url: {}'.format(url))
        res = self.request(url).json()
        self.check_jsonld(res, jsonld_context)

    @parameters(bts.ANNOTATION_GET_MSGPACK)
    def test_annotation_GET_msgpack(self, bid):
        ''' Test that the annotation GET endpoint correctly returns msgpack format. '''
        url = bts.ANNOTATION_ENDPOINT + '/' + bid
        # Test must be set up correctly
        self.assertRegexpMatches(
            bid, r'msgpack=((1)|([Tt][Rr][Uu][Ee]))',
            'Test setup error, "msgpack" parameter not specified or false in url: {}\n'.format(url))
        # Get object without msgpack
        total_url = re.sub(r'[?&]msgpack=((1)|([Tt][Rr][Uu][Ee]))', '', url)
        # get the total and msgpack results
        res_t = self.request(total_url).json()
        res_msgpack = self.convert_msgpack(self.msgpack_ok(self.request(url).content))
        # remove took from both (should be different)
        res_t.pop('took', None)
        res_msgpack.pop('took', None)
        self.assertDictEqual(
            res_t,
            res_msgpack,
            "msgpacked results and non-msgpacked results differ for query: {}".format(url))
    """
    def test_annotation_GET(self):
        ''' Function to test GETs to the annotation endpoint.

            Currently supports automatic testing of:
                fields
                filter
                jsonld
                callback
        '''
        # Check some ids to make sure the resulting _id matches

        for (test_number, bid) in enumerate(ns.annotation_GET):
            base_url = ns.annotation_endpoint + '/' + bid
            get_url = ns.annotation_endpoint + '/' + _q(bid.split('?')[0]) + '?' + '?'.join(bid.split('?')[1:])
            # if it specifies a callback function, make sure it works
            if self.parse_url(base_url, 'callback'):
                res = self.extract_results_from_callback(get_url)
            else:
                res = self.request(get_url).json()
            # Check that the returned ID matches
            eq_(res['_id'], bid.split('?')[0])
            # Is this a jsonld query?
            if self.check_boolean_url_option(base_url, 'jsonld') and 'root' in jsonld_context:
                self.check_jsonld(res, '', jsonld_context)
            if 'fields' in bid or 'filter' in bid:
                # This is a filter query, test it appropriately.  First get a list of fields the user specified
                if 'fields' in bid:
                    true_fields = [x.strip() for x in self.parse_url(base_url, 'fields').split(',')]
                elif 'filter' in bid:
                    true_fields = [x.strip() for x in self.parse_url(base_url, 'filter').split(',')]
                # Next get the object with no fields specified
                total_url = ns.annotation_endpoint + '/' + _q(bid.split('?')[0])
                res_total = self.request(total_url).json()
                # Check the fields
                self.check_fields(res, res_total, true_fields, ns.additional_fields_for_check_fields_subset)

            # override to add more tests
            self._extra_annotation_GET(bid, res)

        # insert non ascii characters
        self.request('' + ns.annotation_endpoint + '/' + '\xef\xbf\xbd\xef\xbf\xbd', expect_status=404)

        # test msgpack on first ID
        self.check_msgpack(self.api + '/' + ns.annotation_endpoint + '/' + _q(ns.annotation_GET[0].split('?')[0]))

        # test unicode string handling
        self.request('' + ns.annotation_endpoint + '/' + ns.unicode_test_string, expect_status=404)

        # test empties
        self.request('' + ns.annotation_endpoint, expect_status=404)
        self.request('' + ns.annotation_endpoint + '/', expect_status=404)


        def test_annotation_POST(self):
        ''' Function to test POSTs to the annotation endpoint.

            Currently supports automatic testing of:
                ids
                fields
                filters
                jsonld
        '''
        base_url = ns.annotation_endpoint
        for (test_number, ddict) in enumerate(ns.annotation_POST):
            res = self.json_ok(self.post_ok(base_url, ddict))
            returned_ids = [h['_id'] if '_id' in h else h['query'] for h in res]
            assert set(returned_ids) == set([x.strip() for x in ddict['ids'].split(',')]), "Set of returned ids doesn't match set of requested ids for annotation POST"
            # Check that the number of returned objects matches the number of inputs
            # Probably not needed given the previous test
            eq_(len(res), len(ddict['ids'].split(',')))
            for hit in res:
                # If it's a jsonld query, check that
                if 'jsonld' in ddict and ddict['jsonld'].lower() in ['true', 1]:
                    self.check_jsonld(hit, '', jsonld_context=jsonld_context)
                # If its a filtered query, check the return objects fields
                if 'filter' in ddict or 'fields' in ddict:
                    true_fields = []
                    if 'fields' in ddict:
                        true_fields = [f.strip() for f in ddict.get('fields').split(',')]
                    elif 'filter' in ddict:
                        true_fields = [f.strip() for f in ddict.get('filter').split(',')]
                    total_url = base_url + '/' + _q(hit['_id'])
                    res_total = self.request(total_url).json()
                    self.check_fields(hit, res_total, true_fields, ns.additional_fields_for_check_fields_subset)

                self._extra_annotation_POST(ddict, hit)

        # Check unicode handling on first test
        res_empty = self.json_ok(self.post_ok(base_url, {'ids': ns.unicode_test_string}))
        assert (len(res_empty) == 1) and (res_empty[0]['notfound']), "POST to annotation endpoint failed with unicode test string"
        # Check unicode test string as the second id in the list
        res_second_empty = self.json_ok(self.post_ok(base_url, {'ids': ns.annotation_POST[0]['ids'].split(',')[0] + ',' + ns.unicode_test_string}))
        assert (len(res_second_empty) == 2) and (res_second_empty[1]['notfound']), "POST to annotation endpoint failed with unicode test string"

    def test_query_GET(self):
        ''' Function to test GETs to the query endpoint.

            Automatically tested parameters:

            fields
            filters
            size
            fetch_all/scroll_id
            facets
            callback

            Separately tested:

            from
        '''
        # Test some simple GETs to the query endpoint, first check some queries to make sure they return some hits
        for (test_number, q) in enumerate(ns.query_GET):
            base_url = ns.query_endpoint + '?q=' + q
            # parse callback
            if self.parse_url(base_url, 'callback'):
                res = self.extract_results_from_callback(base_url)
            elif self.check_boolean_url_option(base_url, 'fetch_all'):
                # Is this a fetch all query?
                # TODO:  make this less crappy.
                sres = self.request(base_url).json()
                assert '_scroll_id' in sres, "_scroll_id not found for fetch_all query: {}".format(q)
                scroll_hits = int(sres['total']) if int(sres['total']) <= 1000 else 1000
                res = self.request(self.api + '/' + ns.query_endpoint + '?scroll_id=' + sres['_scroll_id']).json()
                assert 'hits' in res, "No hits found for query: {}\nScroll ID: {}".format(q, sres['_scroll_id'])
                assert len(res['hits']) == scroll_hits, "Expected a scroll size of {}, got a scroll size of {}".format(scroll_hits, len(res['hits']))
            else:
                # does this query have hits?
                res = self.query_has_hits(q, ns.query_endpoint)
            # Test the size/size cap
            total_hits = int(res['total'])
            ret_size = len(res.get('hits', []))
            req_size = int(self.parse_url(base_url, 'size')) if self.parse_url(base_url, 'size') else 10
            if self.check_boolean_url_option(base_url, 'fetch_all'):
                true_size = scroll_hits
            elif total_hits < req_size:
                true_size = total_hits
            else:
                true_size = req_size if req_size <= 1000 else 1000
            assert ret_size == true_size, 'Expected {} hits for query "{}", got {} hits instead'.format(true_size, q, ret_size)
            # Test facets, maybe we should make it a subset test, i.e., set(returned facets) must be a subset of set(requested facets)
            if 'facets' in q:
                facets = [x.strip() for x in self.parse_url(base_url, 'facets').split(',')]
                assert 'facets' in res, "Facets were expected in the response object, but none were found."
                for facet in facets:
                    assert facet in res['facets'], 'Expected facet "{}" in response object, but it was not found'.format(facet)
            # Exhaustively test the first 10?
            for hit in res['hits'][:10]:
                # Make sure correct jsonld is in res
                if self.check_boolean_url_option(base_url, 'jsonld') and 'root' in jsonld_context:
                    hit = self.check_jsonld(hit, '', jsonld_context)
                if self.parse_url(base_url, 'fields') or self.parse_url(base_url, 'filter'):
                    # This is a filter query, test it appropriately
                    if 'fields' in q:
                        true_fields = [x.strip() for x in self.parse_url(base_url, 'fields').split(',')]
                    elif 'filter' in bid:
                        true_fields = [x.strip() for x in self.parse_url(base_url, 'filter').split(',')]
                    total_url = ns.annotation_endpoint + '/' + _q(hit['_id'])
                    res_total = {}
                    res_total = self.request(total_url).json()
                    if res_total:
                        self.check_fields(hit, res_total, true_fields, ns.additional_fields_for_check_fields_subset)
            # extra tests
            self._extra_query_GET(q, res)

        # test non-ascii characters
        res_f = self.request(self.api + '/' + ns.query_endpoint + '?q=' + '\xef\xbf\xbd\xef\xbf\xbd').json()
        assert res_f['hits'] == [], 'Query with non ASCII characters injected failed'

        # test msgpack on first query only
        self.check_msgpack(self.api + '/' + ns.query_endpoint + '?q=' + ns.query_GET[0])

        # test unicode insertion
        res = self.request(self.api + '/' + ns.query_endpoint + '?q=' + ns.unicode_test_string).json()
        assert res['hits'] == [], "GET to query endpoint failed with unicode test string"

        # test empty/error
        res = self.request(self.api + '/' + ns.query_endpoint), checkerror=False)
        assert 'error' in res, "GET to query endpoint failed with empty query"

    def test_query_POST(self):
        '''
            Test POSTS to the query endpoint.

        '''
        #query via post
        base_url = ns.query_endpoint
        for (test_number, ddict) in enumerate(ns.query_POST):
            res = self.json_ok(self.post_ok(base_url, ddict).json()
            returned_queries = [h['query'] for h in res]
            assert set(returned_queries) == set([x.strip() for x in ddict['q'].split(',')]), "Set of returned queries doesn't match set of requested queries for annotation POST"
            # Check that the number of returned objects matches the number of inputs
            # Probably not needed given the previous test
            eq_(len(res), len(ddict['q'].split(',')))
            for hit in res:
                # If it's a jsonld query, check that
                if 'jsonld' in ddict and ddict['jsonld'].lower() in ['true', 1]:
                    self.check_jsonld(hit, '', jsonld_context)
                # If its a filtered query, check the return objects fields
                if 'filter' in ddict or 'fields' in ddict:
                    true_fields = []
                    if 'fields' in ddict:
                        true_fields = [f.strip() for f in ddict.get('fields').split(',')]
                    elif 'filter' in ddict:
                        true_fields = [f.strip() for f in ddict.get('filter').split(',')]
                    total_url = ns.annotation_endpoint + '/' + _q(hit['_id'])
                    res_total = self.request(total_url).json()
                    self.check_fields(hit, res_total, true_fields, ns.additional_fields_for_check_fields_subset)

                self._extra_query_POST(ddict, hit)

        # test unicode insertion, make sure scopes points to a field that is string indexed or else these will fail...
        res = self.json_ok(self.post_ok(base_url, {'q': ns.unicode_test_string, 'scopes': ns.query_POST[0]['scopes']}), checkerror=False)
        assert (len(res) == 1) and (res[0]['notfound']), "POST to query endpoint failed with unicode test string"

        res = self.json_ok(self.post_ok(base_url, {'q': ns.query_POST[0]['q'].split(',')[0] + ',' +
                                ns.unicode_test_string, 'scopes': ns.query_POST[0]['scopes']}), checkerror=False)
        assert (len(res) == 2) and (res[1]['notfound']), "POST to query endpoint failed with unicode test string"

        # test empty post
        res = self.json_ok(self.post_ok(base_url, {'q': ''}), checkerror=False)
        assert 'error' in res, "POST to query endpoint failed with empty query"
    """

    def test_metadata(self):
        self.request('/metadata')
        self.request('metadata')

    def test_get_fields(self):
        res = self.request(self.api + '/metadata/fields').json()
        # Check to see if there are enough keys
        self.assertGreater(len(res), bts.MINIMUM_NUMBER_OF_ACCEPTABLE_FIELDS,
                           'Too few fields returned in the /metadata/fields endpoint')

        for field in bts.TEST_FIELDS_GET_FIELDS_ENDPOINT:
            self.assertIn(
                field,
                res,
                '"{}" expected in response from /metadata/fields, but not found'.format(field))

    def test_status_endpoint(self):
        self.request('/status')
        # (testing failing status would require actually loading tornado app from there
        #  and deal with config params...)

    @classmethod
    def suite(cls):
        tests = unittest.defaultTestLoader.getTestCaseNames(cls)
        return unittest.TestSuite(map(cls, tests))



import json
from collections import OrderedDict
import datetime
import tornado.web
import re
from biothings.utils.web.analytics import GAMixIn
from biothings.utils.web.tracking import StandaloneTrackingMixin
from biothings.utils.common import is_str, is_seq, split_ids
from urllib.parse import urlencode, urlparse, urlunparse, parse_qs, unquote_plus

try:
    from raven.contrib.tornado import SentryMixin
except ImportError:
    # dummy class mixin
    class SentryMixin(object):
        pass

try:
    from re import fullmatch as match
except ImportError:
    from re import match
import logging

try:
    import msgpack

    def msgpack_encode_datetime(obj):
        if isinstance(obj, datetime.datetime):
            return {'__datetime__': True, 'as_str': obj.strftime("%Y%m%dT%H:%M:%S.%f")}
        return obj
    SUPPORT_MSGPACK = True
except ImportError:
    SUPPORT_MSGPACK = False

try:
    import yaml
    SUPPORT_YAML = True
except ImportError:
    SUPPORT_YAML = False

class DateTimeJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        else:
            return super(DateTimeJSONEncoder, self).default(obj)

class BiothingParameterTypeError(Exception):
    pass

class BaseHandler(SentryMixin, tornado.web.RequestHandler, GAMixIn, StandaloneTrackingMixin):
    ''' Parent class of all biothings handlers, only direct descendant of
        `tornado.web.RequestHandler <http://www.tornadoweb.org/en/stable/web.html#tornado.web.RequestHandler>`_, 
        contains the common functions in the biothings handler universe:

            * return `self` as JSON
            * set CORS and caching headers
            * typify the URL keyword arguments
            * optionally send tracking data to google analytics and integrate with sentry monitor'''

    def initialize(self, web_settings):
        """ Tornado handler `initialize() <http://www.tornadoweb.org/en/stable/web.html#tornado.web.RequestHandler.initialize>`_, 
        Override to add settings for *this* biothing API.  Assumes that the ``web_settings`` kwarg exists in APP_LIST """
        self.web_settings = web_settings
        self.ga_event_object_ret = {'category': '{}_api'.format(self.web_settings.API_VERSION)}
        self.kwarg_settings = {}

    def _format_log_exception_message(self, msg='', delim="-"*30):
        return "{msg}\n\nError message:\n{delim}\n{msg}\n\nRequest parameters:\n{delim}\n{req}\n\nTraceback:\n{delim}\n".format(msg=msg, delim=delim, req=self.request)

    def log_exceptions(self, exception_msg=''):
        """ Logs the current exception in tornado logs.
            This must be called in an exception handler """
        _msg = self._format_log_exception_message(exception_msg)
        logging.exception(_msg)

    def ga_event_object(self, data={}):
        ''' Create the data object for google analytics tracking. '''
        # Most of the structure of this object is formed during self.initialize
        if data:
            self.ga_event_object_ret['label'] = list(data.keys()).pop()
            self.ga_event_object_ret['value'] = list(data.values()).pop()
        return self.ga_event_object_ret

    def _sanitize_params(self, args):
        ''' Subclass to implement custom parameter sanitization '''
        self.jsonp = args.pop(self.web_settings.JSONP_PARAMETER, None)
        return args

    def _typify(self, arg, argval, json_list_input=False):
        ''' Try to get the parameter's type from settings '''
        # first see if this parameter has an alias
        # do value translations, if they exist
        if 'type' not in self.kwarg_settings[arg]:
            return argval

        if self.kwarg_settings[arg]['type'] == list:
            ret = []
            if json_list_input:
                try:
                    ret = json.loads(argval)
                    if not isinstance(ret, list):
                        raise ValueError
                except Exception:
                    ret = []
                    #raise BiothingParameterTypeError('Could not listify "{}" in parameter "{}" with "jsoninput" True'.format(argval, arg))
            if not ret:
                try:
                    ret = split_ids(argval)
                except ValueError as e:
                    raise BiothingParameterTypeError("Could not split argument '{0}' due to the following error: '{1}'".format(arg, str(e))) 
                #ret = [x for x in re.split(getattr(self.web_settings, 'LIST_SPLIT_REGEX', '[\s\r\n+|,]+'), argval) if x]
            ret = ret[:self.kwarg_settings[arg].get('max', getattr(self.web_settings, 'LIST_SIZE_CAP', 1000))]
        elif self.kwarg_settings[arg]['type'] == int:
            try:
                ret = int(argval)
            except ValueError:
                raise BiothingParameterTypeError("Expected '{0}' parameter to have integer type.  Couldn't convert '{1}' to integer".format(arg, argval))
        elif self.kwarg_settings[arg]['type'] == float:
            try:
                ret = float(argval)
            except ValueError:
                raise BiothingParameterTypeError("Expected '{0}' parameter to have float type.  Couldn't convert '{1}' to float".format(arg, argval))
        elif self.kwarg_settings[arg]['type'] == bool:
            ret = self._boolify(argval)
        else:
            ret = argval
        return ret

    def _boolify(self, val):
        return val.lower() in ['1', 'true', 'y', 't']

    def _translate_input_arg_value(self, arg, argval):
        if 'translations' in self.kwarg_settings[arg]:
            for (regex, translation) in self.kwarg_settings[arg]['translations']:
                argval = re.sub(regex, translation, argval)
        return argval 

    def _translate_and_typify_arg_values(self, args, json_list_input=False):
        for _arg in list(args.keys()):
            if _arg in self.kwarg_settings:
                args[_arg] = self._translate_input_arg_value(_arg, args[_arg])
                args[_arg] = self._typify(_arg, args[_arg], json_list_input=json_list_input)
        return args
    
    def _alias_input_args(self, args):
        alias_dict = dict([(_arg, _setting['alias']) for (_arg, _setting) in self.kwarg_settings.items() 
                            if 'alias' in _setting])
        for (target, src) in alias_dict.items():
            if is_str(src) and src in args:
                args.setdefault(target, args[src])
            elif is_seq(src):
                for param in src:
                    if param in args:
                        args.setdefault(target, args[param])
                        break
        return args

    def get_query_params(self):
        '''Extract, typify, and sanitize the parameters from the URL query string. '''
        _args = dict([(k, self.get_argument(k)) for k in self.request.arguments])
        _args = self._alias_input_args(_args)
        _args = self._translate_and_typify_arg_values(_args, json_list_input=self._boolify(_args.get('jsoninput','')))
        _args = self._sanitize_params(_args)
        return _args

    def return_html(self, data, status_code=200):
        self.set_status(status_code)
        if not self.web_settings.DISABLE_CACHING:
            #get etag if data is a dictionary and has "etag" attribute.
            etag = data.get('etag', None) if isinstance(data, dict) else None
            self.set_cacheable(etag=etag)
        self.support_cors()
        self.set_header("Content-Type", "text/html; charset=utf-8")
        _link = self.request.full_url()
        d = urlparse(_link)
        if 'metadata' in d.path:
            _docs = self.web_settings.METADATA_DOCS_URL
        elif 'query' in d.path:
            _docs = self.web_settings.QUERY_DOCS_URL
        else:
            _docs = self.web_settings.ANNOTATION_DOCS_URL
        q = parse_qs(d.query)
        q.pop('format', None)
        d = d._replace(query=urlencode(q, True))
        _link = urlunparse(d)
        self.write(self.web_settings.HTML_OUT_TEMPLATE.format(data=json.dumps(data), 
            img_src=self.web_settings.HTML_OUT_HEADER_IMG, link=_link, link_decode=unquote_plus(_link),
            title_html=self.web_settings.HTML_OUT_TITLE, docs_link=_docs))
        return

    def return_yaml(self, data, status_code=200):

        if not SUPPORT_YAML:
            self.set_status(500)
            self.write({"success":False,"error":"Extra requirements for biothings.web needed."})
            return

        def ordered_dump(data, stream=None, Dumper=yaml.Dumper, **kwds):
            class OrderedDumper(Dumper):
                pass

            def _dict_representer(dumper, data):
                return dumper.represent_mapping(
                    yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                    data.items())
            OrderedDumper.add_representer(OrderedDict, _dict_representer)
            return yaml.dump(data, stream, OrderedDumper, **kwds)
        
        self.set_status(status_code)
        self.set_header("Content-Type", "text/x-yaml; charset=UTF-8")
        self.support_cors()
        self.write(ordered_dump(
            data=data, Dumper=yaml.SafeDumper, default_flow_style=False))

    def return_object(self, data, encode=True, indent=None, status_code=200, _format='json'):
        '''Return passed data object as the proper response.
            
        :param data: object to return as JSON
        :param encode: if encode is False, assumes input data is already a JSON encoded string.
        :param indent: number of indents per level in JSON string
        :param status_code: HTTP status code for response
        :param _format: output format - currently supports "json", "html", "yaml", or "msgpack"
        '''
        if _format == 'html':
            self.return_html(data=data, status_code=status_code)
            return
        elif  _format == 'yaml':
            self.return_yaml(data=data, status_code=status_code)
            return
        elif SUPPORT_MSGPACK and self.web_settings.ENABLE_MSGPACK and _format == 'msgpack':
            self.return_json(data=data, encode=encode, indent=indent, status_code=status_code, is_msgpack=True)
            return
        else:
            self.return_json(data=data, encode=encode, indent=indent, status_code=status_code)

 
    def return_json(self, data, encode=True, indent=None, status_code=200, is_msgpack=False):
        '''Return passed data object as JSON response.
        If **jsonp** parameter is set in the  request, return a valid 
        `JSONP <https://en.wikipedia.org/wiki/JSONP>`_ response.
            
        :param data: object to return as JSON
        :param encode: if encode is False, assumes input data is already a JSON encoded string.
        :param indent: number of indents per level in JSON string
        :param status_code: HTTP status code for response
        :param is_msgpack: should this object be compressed before return?
        '''    
        indent = indent or 2   # tmp settings
        self.set_status(status_code)
        if is_msgpack:
            _json_data = msgpack.packb(data, use_bin_type=True, default=msgpack_encode_datetime)
            self.set_header("Content-Type", "application/x-msgpack")
        else:
            _json_data = json.dumps(data, cls=DateTimeJSONEncoder, indent=indent) if encode else data
            self.set_header("Content-Type", "application/json; charset=UTF-8")
        if not self.web_settings.DISABLE_CACHING:
            #get etag if data is a dictionary and has "etag" attribute.
            etag = data.get('etag', None) if isinstance(data, dict) else None
            self.set_cacheable(etag=etag)
        self.support_cors()
        if getattr(self, 'jsonp', False):
            self.write('%s(%s)' % (self.jsonp, _json_data))
        else:
            self.write(_json_data)

    def set_cacheable(self, etag=None):
        '''set proper header to make the response cacheable.
           set etag if provided.
        '''
        self.set_header("Cache-Control", "max-age={}, public".format(self.web_settings.CACHE_MAX_AGE))
        if etag:
            self.set_header('Etag', etag)

    def support_cors(self, *args, **kwargs):
        '''Provide server side support for CORS request.'''
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Methods", "{}".format(self.web_settings.ACCESS_CONTROL_ALLOW_METHODS))
        self.set_header("Access-Control-Allow-Headers",
                        "{}".format(self.web_settings.ACCESS_CONTROL_ALLOW_HEADERS))
        self.set_header("Access-Control-Allow-Credentials", "false")
        self.set_header("Access-Control-Max-Age", "60")

    def options(self, *args, **kwargs):
        self.support_cors()



from biothings.web.api.es.query import ESQuery as DefaultESQuery
from biothings.web.api.es.query_builder import ESQueryBuilder as DefaultESQueryBuilder
from biothings.web.api.es.transform import ESResultTransformer as DefaultESResultTransformer
from biothings.web.templates import HTML_OUT_TEMPLATE
import re

# *****************************************************************************
# Elasticsearch variables
# *****************************************************************************
# elasticsearch server transport url
ES_HOST = 'localhost:9200'
# timeout for python es client (global request timeout)
ES_CLIENT_TIMEOUT = 120
# elasticsearch index name
ES_INDEX = 'mybiothing_current'
# elasticsearch document type
ES_DOC_TYPE = 'biothing'
# Amount of time a scroll request is kept open
ES_SCROLL_TIME = '1m'
# Size of each scroll request return
ES_SCROLL_SIZE = 1000
# Maximum size of result return
ES_SIZE_CAP = 1000
# Maximum result window => maximum for "from" parameter
ES_RESULT_WINDOW_SIZE_CAP = 10000

# For the userquery folder for this app
USERQUERY_DIR = ''

# default static path, relative to current working dir
# (from where app is launched)
STATIC_PATH = "static"

# api version in the URL patterns and elsewhere
API_VERSION = 'v1'
# for URLS that don't require settings in initialize
UNINITIALIZED_APP_LIST = []
# project URL routing
APP_LIST = []

# Allow the __any__ random doc retrieval
ALLOW_RANDOM_QUERY = False
# Allow facets to be nested with ( )
ALLOW_NESTED_AGGS = False

# *****************************************************************************
# Subclass of biothings.web.api.es.query_builder.ESQueryBuilder to build
# queries for this app
# *****************************************************************************
ES_QUERY_BUILDER = DefaultESQueryBuilder
# *****************************************************************************
# Subclass of biothings.web.api.es.query.ESQuery to execute queries for this app
# *****************************************************************************
ES_QUERY = DefaultESQuery
# *****************************************************************************
# Subclass of biothings.web.api.es.transform.ESResultTransformer to transform
# ES results for this app
# *****************************************************************************
ES_RESULT_TRANSFORMER = DefaultESResultTransformer

OUTPUT_KEY_ALIASES = {}
#OUTPUT_KEY_ALIASES = {'cadd':'schmadd', 'cadd/gene/ccds_id': 'cces_id'}

# Global default cap for list inputs
LIST_SIZE_CAP = 1000

# For Returning a custom message in 4xx responses (or really any response)
ID_REQUIRED_MESSAGE = 'ID required'
ID_NOT_FOUND_TEMPLATE = "ID '{bid}' not found"

# A list of fields to exclude from metadata/fields endpoint
AVAILABLE_FIELDS_EXCLUDED = ['all']

# A path to the available fields notes 
AVAILABLE_FIELDS_NOTES_PATH = ''

# TODO: Need to describe
#ANNOTATION_ID_REGEX_LIST = [(re.compile(r'rs[0-9]+', re.I), 'dbsnp.rsid')]
ANNOTATION_ID_REGEX_LIST = []

# USERQUERY KWARGS

# regex to identify a userquery arg
USERQUERY_KWARG_REGEX = re.compile(r'^uq_\w+$')
# transform to use on the userquery arg
USERQUERY_KWARG_TRANSFORM = lambda x: x[3:]

# kwargs passed into Elasticsearch.get() for /status endpoint
STATUS_CHECK = {
    'id': '',
    'index': '',
    'doc_type': ''
}

# string used in headers to support CORS
ACCESS_CONTROL_ALLOW_METHODS = 'GET,POST,OPTIONS'
ACCESS_CONTROL_ALLOW_HEADERS = 'Content-Type, Depth, User-Agent, X-File-Size, X-Requested-With, If-Modified-Since, X-File-Name, Cache-Control'

# These kwarg descriptions are used in generating a swagger API spec
KWARG_DESCRIPTIONS = {
    '_source': {'name': 'fields', 'text_template': 'a comma-separated list of fields (in dotfield notation) used to limit the fields returned from the matching {biothing_object} hit(s). The supported field names can be found from any {biothing_object} object or from the /metadata/fields endpoint. If "fields=all", all available fields will be returned.{param_type}{param_default_value}{param_max}'},

    'size': {'name': 'size', 'text_template': 'the maximum number of matching {biothing_object} hits to return per batch.{param_type}{param_default_value}{param_max}'},

    'from': {'name': 'from', 'text_template': 'the number of matching {biothing_object} hits to skip, starting from 0.  This can be useful for paging in combination with the "size" parameter.{param_type}{param_default_value}{param_max}'},

    'sort': {'name': 'sort', 'text_template':'the comma-separated list of fields to sort on. Prefix each with "-" for descending order, otherwise in ascending order. Default: sort by descending score.' },

    'dotfield': {'name': 'dotfield', 'text_template': 'control the format of the returned {biothing_object} object. If "true" or "1", all fields will be collapsed into a single level deep object (all nested objects will be a single level deep, using dotfield notation to signify the nested structure){param_type}{param_default_value}{param_max}'},

    'callback': {'name': 'callback', 'text_template': 'you can pass a "callback" parameter to make a JSONP call. Type: string.'},

    'email': {'name': 'email', 'text_template': 'If you are regular users of our services, we encourage you to provide us with an email, so that we can better track the usage or follow up with you.'},

    'out_format': {'name': 'format', 'text_template': 'controls output format of server response, currently supports: "json", "jsonld", "html".{param_type}{param_default_value}{param_max}'},

    'aggs': {'name': 'facets', 'text_template': 'a comma-separated list of fields to return facets on.  In addition to query hits, the fields notated in "facets" will be aggregated by value and bucklet counts will be displayed in the "facets" field of the response object.{param_type}{param_default_value}{param_max}'},

    'facet_size': {'name': 'facet_size', 'text_template': 'the number of facet buckets to return in the response.{param_type}{param_default_value}{param_max}'},
    
    'ids': {'name': 'ids', 'text_template': 'multiple {biothing_object} ids separated by comma. Note that currently we only take the input ids up to 1000 maximum, the rest will be omitted.{param_type}{param_default_value}{param_max}'},

    'q': {'name': 'q', 'text_template': 'Query string.  The detailed query syntax can be found from our [docs]{doc_query_syntax_url}'},

    'scopes': {'name': 'scopes', 'text_template': 'a comma-separated list of fields as the search "scopes" (fields to search through for query term). The available "fields" that can be passed to the "scopes" parameter are listed in the **/metadata/fields** endpoint.{param_type} Default: "scopes=_id".{param_max}'},

    'search': {'name': 'search', 'text_template': 'Pass a search term to filter the available fields.{param_type}{param_default_value}{param_max}'},

    'prefix': {'name': 'prefix', 'text_template': 'Pass a prefix string to filter the available fields.{param_type}{param_default_value}{param_max}'}
}
# Keyword Argument Control
# 
# These parameters control which input kwargs go to which kwarg group for each
# endpoint and operation (e.g. query GET, annotation POST, etc).
# This allows explicit grouping of parameters into inputs of the pipeline
# section that they are germane to.
# CONTROL_KWARGS - general category, used for handler parameters (e.g. raw, rawquery)
# ES_KWARGS go directly to the ESQuery function (e.g. fields, size...)
# ESQB_KWARGS are used to instantiate a query builder class
# TRANSFORM_KWARGS are used to instantiate a result transformer class
# 
# keys are parameter names, values are default values.  If value == None, no
# default is inserted.

# For annotation GET endpoint
ANNOTATION_GET_CONTROL_KWARGS = {'raw': {'default': False, 'type': bool}, 
                                 'rawquery': {'default': False, 'type': bool},
                                 'out_format': {'default': 'json', 'type': str, 'alias':'format'}}
ANNOTATION_GET_ES_KWARGS = {'_source': {'default': None, 'type': list, 'max': 1000, 'alias': ['fields', 'filter']}}
ANNOTATION_GET_ESQB_KWARGS = {}
ANNOTATION_GET_TRANSFORM_KWARGS = {'dotfield': {'default': False, 'type': bool}, 
                                   'jsonld': {'default': False, 'type': bool},
                                   '_sorted': {'default': True, 'type': bool},
                                    'always_list': {'default': [], 'type': list, 'max': 1000},
                                    'allow_null': {'default': [], 'type': list, 'max': 1000}}

# For annotation POST endpoint
ANNOTATION_POST_CONTROL_KWARGS = {'raw': {'default': False, 'type': bool},
                                  'rawquery': {'default': False, 'type': bool},
                                  'ids': {'default': None, 'type': list, 'max': 1000},
                                  'out_format': {'default': 'json', 'type': str, 'alias': 'format'}}
ANNOTATION_POST_ES_KWARGS = {'_source': {'default': None, 'type': list, 'max': 1000, 'alias': ['fields', 'filter']}}
ANNOTATION_POST_ESQB_KWARGS = {}
ANNOTATION_POST_TRANSFORM_KWARGS = {'dotfield': {'default': False, 'type': bool},
                                    'jsonld': {'default': False, 'type': bool},
                                    '_sorted': {'default': True, 'type': bool},
                                    'always_list': {'default': [], 'type': list, 'max': 1000},
                                    'allow_null': {'default': [], 'type': list, 'max': 1000}}

# For query GET endpoint
QUERY_GET_CONTROL_KWARGS = {'raw': {'default': False, 'type': bool},
                            'rawquery': {'default': False, 'type': bool},
                            'q': {'default': None, 'type': str, 
                                'translations': [
                                    #(re.compile(r'chr:', re.I), r'chrom:')
                                ]
                            },
                            'scroll_id': {'default': None, 'type': str},
                            'fetch_all': {'default': False, 'type': bool},
                            'out_format': {'default': 'json', 'type': str, 'alias': 'format'}}
QUERY_GET_ES_KWARGS = {'_source': {'default': None, 'type': list, 'max': 1000, 'alias': ['fields', 'filter']},
                       'from': {'default': None, 'type': int, 'alias': 'skip'},
                       'size': {'default': None, 'type': int, 'alias': 'limit'},
                       'explain': {'default': None, 'type': bool},
                       'aggs': {'default': None, 'type': list, 'max': 1000, 'alias': 'facets'},
                       'sort': {'default': None, 'type': list, 'max': 1000}}
QUERY_GET_ESQB_KWARGS = {'fetch_all': {'default': False, 'type': bool},
                         'userquery': {'default': None, 'type': str, 'alias': ['userfilter']},
                         'facet_size': {'default': 10, 'type': int, 'max': 1000}}
QUERY_GET_TRANSFORM_KWARGS = {'dotfield': {'default': False, 'type': bool},
                              'jsonld': {'default': False, 'type': bool},
                              '_sorted': {'default': True, 'type': bool},
                              'always_list': {'default': [], 'type': list, 'max': 1000},
                              'allow_null': {'default': [], 'type': list, 'max': 1000}}

# For query POST endpoint
QUERY_POST_CONTROL_KWARGS = {'q': {'default': None, 'type': list},
                             'raw': {'default': False, 'type': bool},
                             'rawquery': {'default': False, 'type': bool},
                             'out_format': {'default': 'json', 'type': str, 'alias': 'format'}}
QUERY_POST_ES_KWARGS = {'_source': {'default': None, 'type': list, 'max': 1000, 'alias': ['fields', 'filter']},
                        'size': {'default': None, 'type': int}}
QUERY_POST_ESQB_KWARGS = {'scopes': {'default': None, 'type': list, 'max': 1000, 
                            'translations': [

                            ]}}
QUERY_POST_TRANSFORM_KWARGS = {'dotfield': {'default': False, 'type': bool}, 
                               'jsonld': {'default': False, 'type': bool},
                               '_sorted': {'default': True, 'type': bool},
                               'always_list': {'default': [], 'type': list, 'max': 1000},
                               'allow_null': {'default': [], 'type': list, 'max': 1000}}

# For metadata GET endpoint
METADATA_GET_CONTROL_KWARGS = {'out_format': {'default': 'json', 'type': str, 'alias': 'format'}}
METADATA_GET_ES_KWARGS = {}
METADATA_GET_ESQB_KWARGS = {}
METADATA_GET_TRANSFORM_KWARGS = {'dev': {'default': False, 'type': bool}, 
                                 'search': {'default': None, 'type': str},
                                 'prefix': {'default': None, 'type': str}}

LICENSE_TRANSFORM = {
    # "alias" :  "datasource",
    # "dot.field" :  "datasource"
}

# *****************************************************************************
# Google Analytics Settings
# *****************************************************************************

# Google Analytics Account ID
GA_ACCOUNT = ''

# Turn this to True to start google analytics tracking
GA_RUN_IN_PROD = False

# url for google analytics tracker
GA_TRACKER_URL = 'mybiothing.info'
GA_ACTION_QUERY_GET = 'query_get'
GA_ACTION_QUERY_POST = 'query_post'
GA_ACTION_ANNOTATION_GET = 'biothing_get'
GA_ACTION_ANNOTATION_POST = 'biothing_post'

# for standalone instance tracking
STANDALONE_TRACKING_URL = ''
# dictionary with AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY k,v for standalone user AWS IAM
STANDALONE_AWS_CREDENTIALS = {}
# batch size for standalone tracking (sending requests to AWS lambda)
STANDALONE_TRACKING_BATCH_SIZE = 1000

# override with url for specific project
URL_BASE = 'http://mybiothing.info'

# parameter for JSONP
JSONP_PARAMETER = 'callback'

# Caching behavior
# should caching be disabled by default on handlers?
DISABLE_CACHING = False
CACHE_MAX_AGE = 604800

# Sentry project address
SENTRY_CLIENT_KEY = ''

# JSON-LD PATH
JSONLD_CONTEXT_PATH = ''

# Can turn msgpack functionality off here, will still load msgpack module if available, just won't
# use it to compress requests
ENABLE_MSGPACK = True

LIST_SPLIT_REGEX = re.compile('[\s\r\n+|,]+')

DEFAULT_SCOPES = ['_id']

# path to the git repository for the app-specific code, override
APP_GIT_REPOSITORY = '../'

# For format=html
HTML_OUT_HEADER_IMG = "//:0"
HTML_OUT_TITLE = "<p>MyBioThing.info</p>"
METADATA_DOCS_URL = "javascript:;"
QUERY_DOCS_URL = "javascript:;"
ANNOTATION_DOCS_URL = "javascript:;"
